#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <math.h>
#include<X11/keysym.h>

#define N 30       // nombre max de sommet du polygone
#define couleur 99 // 'couleur' pour la structure image

/*
Projet Info Graphique
Par Julien GONY & Didier STRAUS

gcc -o RemplissageRE RemplissageRE.c -L/usr/X11R6/lib -lX11 -lm

gcc RemplissageRE.c -g -o RemplissageRE -L/usr/X11R6/lib -lX11 -lm

03/03/04
*/

// Type
typedef struct {
	int x;
	int y;
} coord;

typedef struct { // rectangle englobant
	int xmin;
	int ymin;
	int xmax;
	int ymax;
} RE;

typedef struct eq_droite {double m; double b;} eq_droite; // equation de droite

// variables
GC gc;
Display *display;
int screen,x1,r;
Window win,root;
int win_larg, win_haut;
Colormap cmap;

// couleur
GC red_gc,green_gc,blue_gc;
XColor red_col,green_col,blue_col;
char red[] = "#FF0000", green[] = "#00FF00", blue[] = "#0000FF";

// image
XImage *xi;
char *buffer = NULL;
XImage *XIma = NULL;

// polygone
int dim=0;
coord poly[N];

RE rect; // rectangle englobant
unsigned long white_pixel, black_pixel;

void Remplissage(unsigned long cc);
void Croix(int x0, int y0);
void Brensenham(int xa, int ya, int xb, int yb, int coul);
void calculRE();
//int interieur(coord p,unsigned long cc);// calclul des intersections de la demi droite (Ne marche pas)
int interieur2(coord p);
int casParticulier(coord p,int i);
eq_droite calcul_droite(coord pa,coord pb);
coord intersection(eq_droite e1,eq_droite e2);


/* ******************************************************************** */

void contact_serveur(void) {
	if ((display = XOpenDisplay(""))==NULL) {
		fprintf(stderr, "can't open display");
		exit(1);
	}
} /* contact_serveur */


/* ******************************************************************** */

void Creat_fenetre(void) {
short border_width;

screen = DefaultScreen(display);
root = RootWindow(display,screen);

white_pixel= WhitePixel(display,screen);
black_pixel= BlackPixel(display,screen);

cmap = DefaultColormap(display,screen);

border_width = 3;

win_larg = DisplayWidth(display,screen)/4;
win_haut = DisplayHeight(display,screen)/4;

win = XCreateSimpleWindow(display,root,0,0,win_larg,win_haut,
	border_width,black_pixel,white_pixel);
gc=DefaultGC(display,screen);

XStoreName(display,win,"Remplissage RE");

XSetWindowColormap(display, win, cmap) ;

// SetForeground(display, gc, black_pixel);

XMapWindow(display,win);

// On alloue donc le buffer qui va contenir les graphiques manuellement a l'aide d'un malloc

    if (buffer=(char *) malloc(win_larg * win_haut * sizeof(char))) {
      perror("malloc failed \n");
    }


    /* Allocation d'une XImage  partir du buffer, plus rapide pour
       le dessin et l'affichage qu'un Pixmap */

    xi=XCreateImage(display, DefaultVisual(display, screen) ,8, ZPixmap ,0 ,(char *) buffer,
    	win_larg,win_haut,8,0);
} /* Creat_fenetre */


/* ******************************************************************** */

void gerer_evenement(){
char *str3="Julien GONY & Didier STRAUS";
int i,dist,level,x0,y0;
long int p1,p2,p3;

int res;
int *pBitmap;
int w, h;
int x, y, idx;

level=0;
XSelectInput(display,win,ExposureMask | KeyPressMask | ButtonPressMask);

for(;;) {
	KeySym keysym;
	XEvent ev;
	XNextEvent(display,&ev);

	switch(ev.type){
		case Expose :
			/* On va afficher des dessins ;) */

//    XPutImage(display, win, gc, xi, 0, 0, 0, 0, win_larg, win_haut);

			break;

		case KeyPress :
			keysym = XLookupKeysym(&ev.xkey, 0);
  
    		if ((keysym >= XK_Left) && (keysym <= XK_Down)) { // appuyer sur une fleche pour quitter
				XClearWindow(display,win);
				XDrawString(display,win,gc,win_larg/5,win_haut/2,str3,strlen(str3));
				XFlush(display);
				sleep(3);
				XDestroyWindow(display,win);
				XCloseDisplay(display);
				exit(0);
			}
			else { // appuyer sur n'importe quelle touche pour arreter la saisie du polygone et le dessiner
				level=1;
				if (level==1) {
					/* on affiche le polygone qu'on a defini */
					for(i=0;i<dim-1;i++) {
			 			XDrawLine(display,win,gc,poly[i].x,poly[i].y,poly[i+1].x,poly[i+1].y);
			 			Brensenham(poly[i].x,poly[i].y,poly[i+1].x,poly[i+1].y,couleur);
			 		}
				 	XDrawLine(display,win,gc,poly[dim-1].x,poly[dim-1].y,poly[0].x,poly[0].y);
					Brensenham(poly[dim-1].x,poly[dim-1].y,poly[0].x,poly[0].y,couleur);
				}
				calculRE(); // on met a jour le RE
				Remplissage(couleur); // on rempli le dessin
				level=dim=0;
			}
			break;

		case ButtonPress : // le bouton de la souris a ete presse
			/* on recupere les coordonnees de la position de la souris (x,y) */
			x0=ev.xbutton.x;
			y0=ev.xbutton.y;

			Croix(x0, y0); // on affiche une petite croix a la position du point

			if (level==0) {
				poly[dim].x = x0;
				poly[dim].y = y0;
				dim++;
			}
			break;

		default :
			fprintf(stderr,"J'ai recu un evt : %d\n",ev.type);
			break;
	} /* switch */
} /* for */

} /* gerer_evts */


/* ******************************************************************** */
// calcul du rectangle englobant

void calculRE() {
	int i;
	
	blue_gc = XCreateGC(display, win, 0, 0);
    XParseColor(display, cmap, blue, &blue_col);
    XAllocColor(display, cmap, &blue_col);
    XSetForeground(display, blue_gc, blue_col.pixel);
	
	// on initialise le RE
	rect.xmin=30000;
	rect.xmax=-30000;
	rect.ymin=30000;
	rect.ymax=-30000;
	
	for(i=0;i<dim;i++) {
		if (poly[i].x<rect.xmin) {
			rect.xmin=poly[i].x;
		}
		if (poly[i].y<rect.ymin) {
			rect.ymin=poly[i].y;
		}
		if (poly[i].x>rect.xmax) {
			rect.xmax=poly[i].x;
		}
		if (poly[i].y>rect.ymax) {
			rect.ymax=poly[i].y;
		}
	}
	
	/* On affiche le Rectangle Englobant (RE) en bleu */
	XDrawLine(display, win, blue_gc, rect.xmin, rect.ymin, rect.xmin, rect.ymax);
	XDrawLine(display, win, blue_gc, rect.xmax, rect.ymin, rect.xmax, rect.ymax);
	XDrawLine(display, win, blue_gc, rect.xmin, rect.ymin, rect.xmax, rect.ymin);
	XDrawLine(display, win, blue_gc, rect.xmax, rect.ymax, rect.xmin, rect.ymax);	
} /* calculRE */


/* ******************************************************************** */

eq_droite calcul_droite(coord pa,coord pb) {
	double dx,dy,m,b;
	eq_droite res;

	dx=pb.x-pa.x;
	dy=pb.y-pa.y;
	m=dy/dx;
	b=pa.y-m*pa.x;

	res.m=m;
	res.b=b;
	/* y = m*x + b */
	return res;
} /* calcul_droite */


/* ******************************************************************** */

coord intersection(eq_droite e1,eq_droite e2) {
	/* on considere que les droites se coupent */
	coord res;

	res.x=(e2.b-e1.b)/(e1.m-e2.m);
	res.y=e1.m*res.x + e1.b;

	/* avant de renvoyer res, il faut verifier que le point d'intersection au segment */

	return res;
} /* intersection */


/************************************************************************/

int interieur3(coord p){
	int nbSegementCoupe=0,i;
	eq_droite eq1,eq2;
	coord pTemp,pIntersect;
	
	for(i=0;i<dim-1;i++) { // tous les segments du poly

		if (((p.y<poly[i+1].y) && (p.y>poly[i].y)) || ((p.y<poly[i].y) && (p.y>poly[i+1].y))) {
		// si intersecte la droite
			eq1= calcul_droite(poly[i],poly[i+1]); // equation du segement
			pTemp.x=0;
			pTemp.y=p.y;
			eq2= calcul_droite(p,pTemp); // equation de la ligne
			pIntersect=intersection(eq1,eq2);//intersection entre la ligne et le segment
			if (pIntersect.x<p.x) { // le pt est sur la 1/2 droite
				nbSegementCoupe++;
			}

		}
		else {
			if ((p.y==poly[i].y) && (p.x>poly[i].x)) { // cas particulier
				nbSegementCoupe+=casParticulier(p,i);
			}
		}
	}
	// pour le segment restant  --> dim-1  0

	if (((p.y<poly[0].y) && (p.y>poly[dim-1].y)) || ((p.y<poly[dim-1].y) && (p.y>poly[0].y))) {
		eq1= calcul_droite(poly[0],poly[dim-1]);
		pTemp.x=0;
		pTemp.y=p.y;
		eq2= calcul_droite(p,pTemp);
		pIntersect=intersection(eq1,eq2);
		if (pIntersect.x<p.x) { // le pt est sur la 1/2 droite
			nbSegementCoupe++;
		}
	}
	else {
		if ((p.y==poly[dim-1].y) && (p.x>poly[dim-1].x)) { // cas particulier
			nbSegementCoupe+=casParticulier(p,dim-1);
		}
	}

	if ((nbSegementCoupe%2)==1) {
		return(1);
	}
	else {
		return(0);
	}
} /* interieur3 */


/* ******************************************************************** */
// si le point appartient a deux droites, verifie si compte double ou simple

int casParticulier(coord p,int i) {

	if (i==0) {
		if (((poly[dim-1].y<p.y) && (poly[1].y<p.y)) || ((poly[dim-1].y>p.y) && (poly[1].y>p.y))) {
			return(2);
		}
		else {
			return(1);
		}
	}
	if (i==dim-1) {
		if (((poly[dim-2].y<p.y) && (poly[0].y<p.y)) || ((poly[dim-2].y>p.y) && (poly[0].y>p.y))) {
			return(2);
		}
		else{
			return(1);
		}
	}

	if (((poly[i-1].y<p.y) && (poly[i+1].y<p.y)) || ((poly[i-1].y>p.y) && (poly[i+1].y>p.y))) {
		return(2);
	}
	else {
		return(1);
	}
} /* casParticulier */


/* ******************************************************************** */

void Remplissage(unsigned long cc) {
	unsigned long cpi; // ?
	int i,j;
	coord p;
	
	green_gc = XCreateGC(display, win, 0, 0);
    XParseColor(display, cmap, green, &green_col);
    XAllocColor(display, cmap, &green_col);
    XSetForeground(display, green_gc, green_col.pixel);
	
	for(i=rect.xmin;i<rect.xmax;i++) {
		for(j=rect.ymin;j<rect.ymax;j++) {

			if (XGetPixel(xi,i,j) != cc) { // pour ne pas effacer les contours du polygone
				p.x=i;
				p.y=j;			
				if (interieur3(p) == 1) {
					XDrawPoint(display,win,green_gc,p.x,p.y);
				}
			}
		}
	}
} /* Remplissage */

/* ******************************************************************** */

// pg principal
int main(void) {
	contact_serveur();
	Creat_fenetre();
	gerer_evenement();

	return 0;
} /* main */



/* ******************************************************************** */

void Croix(int x0,int y0) {
	XDrawPoint(display,win,gc,x0,y0);
	XDrawPoint(display,win,gc,x0+1,y0);
	XDrawPoint(display,win,gc,x0-1,y0);
	XDrawPoint(display,win,gc,x0,y0-1);
	XDrawPoint(display,win,gc,x0,y0+1);
} /* Croix */


/* ******************************************************************** */

void Brensenham(int xa, int ya, int xb, int yb, int coul) {
 int x,y,dx,dy,incrmX,incrmY,dp,NE,SE;
 red_gc = XCreateGC(display, win, 0, 0);
 XParseColor(display, cmap, red, &red_col);
 XAllocColor(display, cmap, &red_col);
 XSetForeground(display, red_gc, red_col.pixel);

 dx = xb-xa;
 dy = yb-ya;

 if (dx>0)
  incrmX = 1;
 else {
  incrmX = -1;
  dx *= -1;
 }

 if (dy>0)
  incrmY = 1;
 else {
  incrmY = -1;
  dy *= -1;
 }

 if (dx>=dy) {
  dp=2*dy-dx;
  SE=2*dy;
  NE=2*(dy-dx);

  y=ya;
  for(x=xa;x!=xb;x+=incrmX) {
   /* affichePixel(x,y,coul) */
   XDrawPoint(display,win,red_gc,x, y);
   XPutPixel(xi, x, y, coul);

   if (dp<=0) /* on choisit le pixel E */
    dp += SE;
   else {
    dp += NE;
    y+=incrmY;
   }
  }
 }
 else if (dx<dy) {
  dp=2*dx-dy;
  SE=2*dx;
  NE=2*(dx-dy);

  x=xa;
  for(y=ya;y!=yb;y+=incrmY) {
   /* affichePixel(x,y,coul) */
   XDrawPoint(display,win,red_gc,x, y);
   XPutPixel(xi, x, y, coul);

   if (dp<=0) /* on choisit le pixel E */
    dp += SE;
   else {
    dp += NE;
    x+=incrmX;
   }
  }
 }
} /* Brensenham */
